
// timer.c implementation file
//----------------------------
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <errno.h>
#include "timer.h"
#include "defines.h"

#define TIMEOUTS		16
#define TIMEOUT_SIGNAL	(SIGRTMIN+0)

#define TIMEOUT_USED	1
#define TIMEOUT_ARMED	2
#define TIMEOUT_PASSED	4

static timer_t	 timeout_timer;
static volatile sig_atomic_t timeout_state[TIMEOUTS]={0};
static struct timespec timeout_time[TIMEOUTS];

static inline double timespec_diff(const struct timespec after,
	const struct timespec before)
//=================================================================
// Return the number of seconds between before and after, (after-before).
// This must be async-signal save, so it cannot use difftime().
//-----------------------------------------------------------------------
{
	return (double)(after.tv_sec-before.tv_sec)
		+(double)(after.tv_nsec-before.tv_nsec)/1000000000.0;
}

static inline void timespec_set(struct timespec* const to,
	const double seconds)
//===========================================================
// Set the timespec to the specified number of seconds,
// or zero if negative seconds.
//-----------------------------------------------------
{
	long sec;
	long nsec;
	if (to)
	{
		if (seconds>0.0)
		{
			sec=(long)seconds;
			nsec=(long)(0.5+1000000000.0*(seconds-(double)sec)); 
			if (nsec<0L)
			{
				nsec=0L;
			}
			else
			{
				if (nsec>999999999.0)
				{
					nsec=999999999.0;
				}
			}
			to->tv_sec=(time_t)sec;
			to->tv_nsec=nsec;
		}
		else
		{
			to->tv_sec=(time_t)0;
			to->tv_nsec=0L;
		}
	}
}

static inline void timespec_add(struct timespec* const to,
	const double seconds)
//===========================================================
// Add positive seconds to a timespec, nothing if seconds is negative.
// This must be async-signal safe.
//--------------------------------------------------------------------
{
	long sec;
	long nsec;	
	if (to && seconds>0.0)
	{
		sec=(long)seconds;
		nsec=(long)(0.5+1000000000.0*(seconds-(double)sec));
		
		// Adjust for rounding errors
		//---------------------------
		if (nsec<0L)
		{
			nsec=0L;
		}
		else
		{
			if (nsec>999999999.0)
			{
				nsec=999999999.0;
			}
		}
		to->tv_sec=(time_t)sec;
		to->tv_nsec=nsec;
		if (to->tv_nsec>=1000000000L)
		{
			to->tv_nsec-=1000000000L;
			to->tv_sec++;
		}
	}
}

static void timeout_signal_handler(int signum __attribute__((unused)),
	siginfo_t* info, void* context __attribute__((unused)))
//=====================================================================
{
	struct timespec now;
	struct itimerspec when;
	int saved_errno;
	int index;
	double next;
	
	// Check for a timer signal
	//-------------------------
	if (!info || info->si_code!=SI_TIMER)
	{
		return;
	}
	
	// Save errno, some of the functions used may modify errno
	//--------------------------------------------------------
	saved_errno=errno;
	if (clock_gettime(CLOCK_REALTIME,&now))
	{
		errno=saved_errno;
		return;
	}
	
	// Assume no next timeout
	//-----------------------
	next=-1.0;
	
	// Check all timeouts that are used and armed, but not passed yet
	//---------------------------------------------------------------
	for (index=0; index<TIMEOUTS; index++)
	{
		if ((__sync_or_and_fetch(&timeout_state[index],0)&(TIMEOUT_USED
			|TIMEOUT_ARMED|TIMEOUT_PASSED))==(TIMEOUT_USED|TIMEOUT_ARMED))
		{
			const double seconds=timespec_diff(timeout_time[index],now);
			if (seconds<=0.0)
			{
				// The timeout[index] fires!
				//--------------------------
				__sync_or_and_fetch(&timeout_state[index],TIMEOUT_PASSED);
			}
			else
			{
				if (next<=0.0 || seconds<next)
				{
					// this is the soonest timeout in the future
					//------------------------------------------
					next=seconds;
				}
			}
		}
	}
	
	// Note: timespec_set() will set the time to zero if next<=0.0,
	// which in turn will disarm the timer.
	// The timer is one-shot, it_interval==0
	//-------------------------------------------------------------
	timespec_set(&when.it_value,next);
	when.it_interval.tv_sec=0;
	when.it_interval.tv_nsec=0L;
	timer_settime(timeout_timer,0,&when,NULL);
	
	// Restore errno
	//--------------
	errno=saved_errno;
}

int timer_init(void)
//===================
{
	struct sigaction act;
	struct sigevent evt;
	struct itimerspec arm;
	
	// Install timeout_signal_handler
	//-------------------------------
	sigemptyset(&act.sa_mask);
	act.sa_sigaction=timeout_signal_handler;
	act.sa_flags=SA_SIGINFO;
	if (sigaction(TIMEOUT_SIGNAL,&act,NULL))
	{
		return errno;
	}
	
	// Create a timerthat willsignal to timeout_signal_handler
	//--------------------------------------------------------
	evt.sigev_notify=SIGEV_SIGNAL;
	evt.sigev_signo=TIMEOUT_SIGNAL;
	evt.sigev_value.sival_ptr=NULL;
	if (timer_create(CLOCK_REALTIME,&evt,&timeout_timer))
	{
		return errno;
	}
	
	// Disarm the timeout timer for now
	//---------------------------------
	arm.it_value.tv_sec=0;
	arm.it_value.tv_nsec=0L;
	arm.it_interval.tv_sec=0;
	arm.it_interval.tv_nsec=0L;
	if (timer_settime(timeout_timer,0,&arm,NULL))
	{
		return errno;
	}
	return OKAY;
}

int timer_passed(const int timeout)
//==================================
// Return nonzero if the timeout has occured
//------------------------------------------
{
	if (timeout>=0 && timeout<TIMEOUTS)
	{
		const int state=__sync_or_and_fetch(&timeout_state[timeout],0);
		
		// Check if it refers to an unused timeout
		//----------------------------------------
		if (!(state&TIMEOUT_USED))
		{
			return ERROR;
		}
		
		// Check if armed
		//---------------
		if (!(state&TIMEOUT_ARMED))
		{
			return ERROR;
		}
		
		// Return >0 if timeout passed, 0 otherwise
		//-----------------------------------------
		if ((state&TIMEOUT_PASSED)!=0)
		{
			return VALID;
		}
		else
		{
			return OKAY;
		}
	}
	else
	{
		// Invalid timeout number
		//-----------------------
		return ERROR;
	}
}

int timer_set(const double seconds)
//===================================
{
	struct timespec now;
	struct timespec then;
	struct itimerspec when;
	double next;
	int timeout;
	int index;
	
	// Timeout must be in the future
	//------------------------------
	if (seconds<=0.0)
	{
		fprintf(stderr,"timer_set: invalid seconds parameter\n");
		return ERROR;
	}
	
	// Get the current time
	//---------------------
	if (clock_gettime(CLOCK_REALTIME,&now))
	{
		fprintf(stderr,"timer_set: clock_gettime failed\n");
		return ERROR;
	}
	
	// Calculate when the timeout should fire
	//---------------------------------------
	then=now;
	timespec_add(&then,seconds);
	
	// Find a unused timeout
	//----------------------
	for (timeout=0; timeout<TIMEOUTS; timeout++)
	{
		if (!(__sync_fetch_and_or(&timeout_state[timeout],TIMEOUT_USED)&TIMEOUT_USED))
		{
			break;
		}
	}
	
	// Check for no unused timeout
	//----------------------------
	if (timeout>=TIMEOUTS)
	{
		fprintf(stderr,"timer_set: no unused timeout\n");
		return ERROR;
	}
	
	// Clear all but TIMEOUT_USED from the state
	//------------------------------------------
	__sync_and_and_fetch(&timeout_state[timeout],TIMEOUT_USED);
	
	// Update the timeout details
	//---------------------------
	timeout_time[timeout]=then;
	
	// Mark the timeout armable
	//-------------------------
	__sync_or_and_fetch(&timeout_state[timeout],TIMEOUT_ARMED);
	
	// How long till the next timeout
	//-------------------------------
	next=seconds;
	for (index=0; index<TIMEOUTS; index++)
	{
		if (__sync_fetch_and_or(&timeout_state[index],0)&(TIMEOUT_USED
			|TIMEOUT_ARMED|TIMEOUT_PASSED)==(TIMEOUT_USED|TIMEOUT_ARMED))
		{
			const double secs=timespec_diff(timeout_time[index],now);
			if (secs>=0.0 && secs<next)
			{
				next=secs;
			}
		}
	}
	
	// Calculate duration when to fire the timeout next
	//-------------------------------------------------
	timespec_set(&when.it_value,next);
	when.it_interval.tv_sec=0;
	when.it_interval.tv_nsec=0L;
	
	// Arm the timer
	//--------------
	if (timer_settime(timeout_timer,0,&when,NULL))
	{
		__sync_and_and_fetch(&timeout_state[timeout],0);
		fprintf(stderr,"timer_set: timer_settime failed\n");
		return ERROR;
	}
	
	// Return the timeout number
	//--------------------------
	return timeout;
}

int timer_unset(const int timeout)
//=================================
{
	if (timeout>=0 && timeout<TIMEOUTS)
	{
		// Obtain the current timeout state to 'state',
		// then clear all but the TIMEOUT_PASSED flag
		// for the specified timeout.
		//---------------------------------------------
		const int state=__sync_fetch_and_and(&timeout_state[timeout],TIMEOUT_PASSED);
		
		// Invalid timeout
		//----------------
		if (!(state&TIMEOUT_USED))
		{
			return ERROR;
		}
		
		// Check if armed
		//---------------
		if (!(state&TIMEOUT_ARMED))
		{
			return ERROR;
		}
		
		// Return 1 if passed, otherwise 0
		//--------------------------------
		if (state&TIMEOUT_PASSED)
		{
			return VALID;
		}
		else
		{
			return OKAY;
		}
	}
	else
	{
		// Invalid timeout number
		//-----------------------
		return ERROR;
	}
}

int timer_done(void)
//===================
{
	struct sigaction act;
	struct itimerspec arm;
	int errors;
	
	// ignore the timeout signals
	//---------------------------
	errors=OKAY;
	sigemptyset(&act.sa_mask);
	act.sa_handler=SIG_IGN;
	if (sigaction(TIMEOUT_SIGNAL,&act,NULL))
	{
		if (!errors)
		{
			errors=errno;
		}
	}
	
	// Disarm any current timeouts
	//----------------------------
	arm.it_value.tv_sec=0;
	arm.it_value.tv_nsec=0L;
	arm.it_interval.tv_sec=0;
	arm.it_interval.tv_nsec=0L;
	if (timer_settime(timeout_timer,0,&arm,NULL))
	{
		if (!errors)
		{
			errors=errno;
		}
	}
	
	// Destroy the timer itself
	//-------------------------
	if (timer_delete(timeout_timer))
	{
		if (!errors)
		{
			errors=errno;
		}
	}
	
	// If any error occured, set errno
	//--------------------------------
	if (errors)
	{
		errno=errors;
	}
	
	// Return 0 if success, errno otherwise
	//-------------------------------------
	return errors;
}
